#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <asm/arch/ns9750_bbus.h>
#include <asm/arch/armadillo3x0_gpio.h>
#include <linux/delay.h>
#include <asm/io.h>

#include "sample_gpio.h"

#ifndef SAMPLE_MINOR
#define SAMPLE_MINOR 185
#endif
#define DRIVER_NAME "sample_gpio_driver"

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Sample driver"
#define LICENSE     "GPL v2"

#define VERSION     DRIVER_NAME ": " DESCRIPTION ", (C) 2007 " AUTHOR

#ifdef DEBUG
#define KDEBUG(args...) printk(KERN_INFO SW_NAME ": " args)
#define KDEBUG_FUNC() KDEBUG("%s()\n", __FUNCTION__)
#else
#define KDEBUG(args...)
#define KDEBUG_FUNC()
#endif

#define INV_NONE   0
#define INV_OUTPUT 1
#define INV_INPUT  2

struct gpio_info {
	int no;
	int gpio;
	int func;
	int inv;
};

static struct gpio_info gpios[] = {
	{ GPIO0, 42, NS_GPIO_CFG_FUNC_GPIO, INV_OUTPUT },
	{ GPIO1, 43, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
	{ GPIO2, 45, NS_GPIO_CFG_FUNC_GPIO, INV_NONE   },
	{ GPIO3, 46, NS_GPIO_CFG_FUNC_GPIO, INV_OUTPUT },
	{ GPIO4, 47, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
	{ GPIO5, 48, NS_GPIO_CFG_FUNC_GPIO, INV_NONE   },
	{ GPIO6, 14, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
	{ GPIO7, 15, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
};

static int
sample_open(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev); 
	KDEBUG_FUNC();
	if(minor != SAMPLE_MINOR)
		return -ENODEV;

	return 0;
}

static int
sample_release(struct inode *inode, struct file *file)
{
	return 0;
}

#define set_gpio_bit(ctl1, ctl2, gpio)          \
({                                              \
	if ((gpio) > 32)                        \
		(ctl2) |= (1 << ((gpio) - 32)); \
	else                                    \
		(ctl1) |= (1 << (gpio));        \
})

static int
sample_ioctl(struct inode *inode, struct file *file,
	     unsigned int cmd, unsigned long arg)
{
	struct sample_arg *larg = (struct sample_arg *)arg;
	spinlock_t lock;
	unsigned long flags;
	int i, j;

	KDEBUG_FUNC();

	//printk("cmd: %d\n", cmd);
	//printk("buf: %08x\n", (unsigned int)larg->buf);
	//printk("size: %d\n", larg->size);

	spin_lock_init(&lock);
	spin_lock_irqsave(&lock, flags);
	for (i=0; i<larg->size; i++) {
		u32 gpio_ctl1 = readl(NS_BBUS_GPIO_CTRL_BASE);
		u32 gpio_ctl2 = readl(NS_BBUS_GPIO_CTRL_BASE + 0x4);
	  
		/*mask*/
		gpio_ctl1 &= ~0x0000c000;
		gpio_ctl2 &= ~0x0001ec00;

		for (j=0; j<8; j++)
			if (larg->buf[i] & (0x1 << j))
				set_gpio_bit(gpio_ctl1, gpio_ctl2,
					     gpios[j].gpio);

		writel(gpio_ctl1, NS_BBUS_GPIO_CTRL_BASE);
		writel(gpio_ctl2,NS_BBUS_GPIO_CTRL_BASE + 4);
	}
	spin_unlock_irqrestore(&lock, flags);

	return 0;
}

static int
gpio_init(void)
{
	int i;

	KDEBUG_FUNC();

	for (i=0; i<ARRAY_SIZE(gpios); i++) {
		if(!request_gpio(gpios[i].gpio, 1, DRIVER_NAME))
			goto error;
        }

	/* default OUTPUT-LOW */
	for (i=0; i<ARRAY_SIZE(gpios); i++){
		set_gpio_cfg(gpios[i].gpio,
			     gpios[i].func | NS_GPIO_CFG_OUTPUT);
		set_gpio_ctrl(gpios[i].gpio, 0);
	}

	return 0;
 error:
	while (i--)
		release_gpio(gpios[i].gpio, 1);

	return -EBUSY;
}

static void
gpio_release(void)
{
	int i;
	KDEBUG_FUNC();
	for (i=0; i<ARRAY_SIZE(gpios); i++){
		set_gpio_cfg(gpios[i].gpio, 
			     gpios[i].func | NS_GPIO_CFG_INPUT);
		release_gpio(gpios[i].gpio, 1);
	}
}

static struct file_operations sample_fops =
{
	owner:		THIS_MODULE,
	ioctl:		sample_ioctl,
	open:		sample_open,
	release:	sample_release,
};

static struct miscdevice sample_miscdev =
{
	SAMPLE_MINOR,
	DRIVER_NAME,
	&sample_fops
};

int __init
sample_init(void)
{
	int ret;
	KDEBUG_FUNC();
	ret = misc_register(&sample_miscdev);
	if (ret < 0) {
		printk("failed register misc\n");
		return ret;
	}

	ret = gpio_init();
	if (ret < 0) {
	  	misc_deregister(&sample_miscdev);
		return ret;
	}

	return 0;
}	

void __exit
sample_exit(void)
{
	KDEBUG_FUNC();
	gpio_release();
	misc_deregister(&sample_miscdev);
}

module_init(sample_init);
module_exit(sample_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);
